#include <jstring.h>
#include <vector>
#include <list>
#include "drawgraph.h"
#include "elttypes.h"
#include <math.h>
#include <auto_arr.h>
#include <assert.h>

// ---------------------- Edge member functions

// Even used for straight lines to aid in depth sorting
void Edge :: GenerateSpline(int degree, Coord Control[][3])
  { // uses detail
    // adds an arc to arcs
    double step=1./setting.detail, t=step;

    arcs.push_back();

    GraphPoint NextPoint;
    NextPoint.p[0]=Control[0][0];
    NextPoint.p[1]=Control[0][1];
    NextPoint.p[2]=Control[0][2];
    arcs.back().push_back(NextPoint);

    // used to be quadratic, degree fixed at 2
    instl auto_array<Coord> c(new Coord[degree+1]);

    for (int i=1; i<=setting.detail; ++i, t+=step)
      {
        if (i<setting.detail)
          {
            Coord oneminust=1-t;
            int j;

            // Coord c[3]={
            //    oneminust*oneminust,
            //    2*t*oneminust,
            //    t*t};
            c[0]=1.;
            for (j=0; j<degree; ++j)
              c[0] *= oneminust;
            for (j=1; j<=degree; ++j)
              c[j] = c[j-1] * t * (degree-j+1) / (oneminust * j);
            //             (d)      d-j  j
            // here c[j] = (j) (1-t)    t

            NextPoint.p[0]=
            NextPoint.p[1]=
            NextPoint.p[2]=0;
            for (j=0; j<=degree; ++j)
              {
                NextPoint.p[0]+=c[j]*Control[j][0];
                NextPoint.p[1]+=c[j]*Control[j][1];
                NextPoint.p[2]+=c[j]*Control[j][2];
              }
          }
        else
          {
            NextPoint.p[0]=Control[degree][0];
            NextPoint.p[1]=Control[degree][1];
            NextPoint.p[2]=Control[degree][2];
          }

        arcs.back().push_back(NextPoint);    
      }
  }

    // Also modifies neighboring vertices to have correct
    // HaloDepth and TextDepth by calling UpdateVertex
void Edge :: Set(int v1, int v2, int multiplicity, Graph &graph)
  {
    assert(multiplicity>=1);
    v[0]=v1;
    v[1]=v2;
    if (v1==v2)
      {
        // worry about which direction
        Coord Away[3];
        int i;
        Coord scale=0;
        for (i=0; i<3; ++i)
          {
            Away[i]=graph.Vert(v1).P[i] - graph.Center[i];
            scale += Away[i]*Away[i];
          }
        if (scale < epsilon)
          {
            Away[0]=1;
            Away[1]=Away[2]=0;
            scale=setting.LoopSize;
          }
        else
          scale = setting.LoopSize / sqrt(scale);

        // make a loop out of a cubic spline
        Coord ControlPts[4][3];
        for (i=0; i<3; ++i)
          {
            ControlPts[0][i]=
            ControlPts[3][i]=graph.Vert(v1).P[i];
          }

        Coord Offset[3];
        Offset[0]=Away[1]+Away[2];
        Offset[1]=0;
        Offset[2]=-Away[0];
        for (int j=0; j<multiplicity; ++j)
          {
            for (i=0; i<3; ++i)
              {
                ControlPts[1][i] = ControlPts[0][i] + (j*.5+1) * scale * (Away[i] + setting.LoopAspect * Offset[i]) ;
                ControlPts[2][i] = ControlPts[0][i] + (j*.5+1) * scale * (Away[i] - setting.LoopAspect * Offset[i]) ;
              }
            
            GenerateSpline(3, ControlPts);  // cubic spline
          }
      }
    else 
      {
        // draw 
        //  (|)
        // between vertices
        Coord ControlPts[3][3], MidPoint[3], Offset[3];
        int i;
        for (i=0; i<3; ++i)
          {
            ControlPts[0][i]=graph.Vert(v1).P[i];
            ControlPts[2][i]=graph.Vert(v2).P[i];
            MidPoint[i]=(ControlPts[0][i]+ControlPts[2][i])/2;
          }

        Offset[0]=(ControlPts[0][1]+ControlPts[0][2]-ControlPts[2][1]-ControlPts[2][2]);
        Offset[1]=0;
        Offset[2]=(ControlPts[2][0]-ControlPts[0][0]);

        Coord step=setting.MultiEdgeWidth/pow(Offset[0]*Offset[0]+Offset[2]*Offset[2], .75); // /norm^1.5
        Coord offmult=step*(multiplicity-1)/2;
        do
          {
            for (i=0; i<3; ++i)
              ControlPts[1][i] = offmult*Offset[i]+MidPoint[i];
            GenerateSpline(2, ControlPts);  // quadratic spline
            offmult -= step;
          } while (--multiplicity);
      }
    // modify neighboring vertices
    ModifyNeighbors(graph);
  }

    // modify neighboring vertices
void Edge :: ModifyNeighbors(Graph &graph)
  {
    for (int whichv=0; whichv< ((v[0]==v[1]) ? 1 : 2); ++whichv)
      for (arciter i=arcs.begin(); i!=arcs.end(); ++i)
        {
          EdgeSplineIter j=(*i).begin();
          assert(j!=(*i).end());
          Coord PrevZ=(*j).p[2];
          bool usenext=false;
          for (; j!=(*i).end(); ++j)
            {
              Coord dist1 = (*j).p[0]-graph.Vert(v[whichv]).P[0];
              dist1*=dist1;
              Coord dist2 = (*j).p[1]+(*j).p[2] 
                            - graph.Vert(v[whichv]).P[1] 
                            - graph.Vert(v[whichv]).P[2];
              dist2*=dist2;
              Coord CurrZ=(*j).p[2];
              if (dist1+dist2 < 144.) // HACK: everything within 12 points 
                {
                  UpdateVertex(whichv, PrevZ, CurrZ, graph);
                  usenext=true;
                }
              else if (usenext)
                {
                  UpdateVertex(whichv, PrevZ, CurrZ, graph);
                  usenext=false;
                }
              PrevZ=CurrZ;
            }
        }
  }

void Edge :: IncreaseMultiplicity(int multiplicity, Graph &graph)
  {
    assert(multiplicity>0);
    this->Set(v[0], v[1], arcs.size()+multiplicity, graph);
  }

void Edge :: UpdateVertex(int whichv, Coord Z1, Coord Z2, Graph &graph)
  {
    whichv=v[whichv];
    Coord minZ=instl min(Z1, Z2);
    Coord maxZ=Z1+Z2-minZ;
    graph.Vert(whichv).HaloDepth = instl max(maxZ+epsilon, graph.Vert(whichv).HaloDepth);
    graph.Vert(whichv).TextDepth = instl min(minZ-epsilon, graph.Vert(whichv).TextDepth);
  }

void Edge :: Render(Scene &scene)
  {
    for (arciter i=arcs.begin(); i!=arcs.end(); ++i)
      {
        Coord PrevPoint[2];
        EdgeSplineIter j=(*i).begin();
        assert(j!=(*i).end());
        Project((*j).p, PrevPoint);
        Coord PrevZ, PrevMZ;
        PrevZ=PrevMZ=(*j).p[2];
        for (++j; j!=(*i).end(); ++j)
          {
            Coord NextPoint[2];

            Project((*j).p, NextPoint);
            Coord Z=(PrevZ+(*j).p[2])/2;
            Coord halodepth=instl max(Z, PrevMZ)+2*epsilon;

            scene << new Disc(PrevPoint, setting.LineRThick, halodepth, true)                 // join halo
                  << new Disc(PrevPoint, setting.LineRNormal, PrevZ)                          // join
                  << new Line(PrevPoint, NextPoint, setting.LineRThick, halodepth, true)      // line halo
                  << new Line(PrevPoint, NextPoint, setting.LineRNormal, Z);                  // line

            PrevPoint[0]=NextPoint[0];
            PrevPoint[1]=NextPoint[1];
            PrevMZ=Z;
            PrevZ=(*j).p[2];
          }
         scene << new Disc(PrevPoint, setting.LineRThick, instl max(PrevMZ,PrevZ)+2*epsilon, true) // end halo
               << new Disc(PrevPoint, setting.LineRThick, PrevZ);                             // end
      }
  }

void Edge :: UpdateBounds(Graph &graph)
  {
    for (arciter i=arcs.begin(); i!=arcs.end(); ++i)
      for (EdgeSplineIter j=(*i).begin(); j!=(*i).end(); ++j)
        for (int k=0; k<3; ++k)
          {
            Coord R=(k==1) ? 0 : setting.LineRThick;
            graph.Bounds[0][k]=instl min((*j).p[k] - R, graph.Bounds[0][k]);
            graph.Bounds[1][k]=instl max((*j).p[k] + R, graph.Bounds[1][k]);
          }
  }

void Edge :: Translate(Coord t[3])
  {
    for (arciter i=arcs.begin(); i!=arcs.end(); ++i)
      for (EdgeSplineIter j=(*i).begin(); j!=(*i).end(); ++j)
        for (int k=0; k<3; ++k)
          (*j).p[k]+=t[k];
  }

